"""Utility functions and classes."""
from enum import Enum
from typing import TypeVar


T = TypeVar('T')

class Registry:
    """
    A registry for managing different types of classes.
    Allows for registering classes with a type name and creating instances of these classes.
    The registry is case-insensitive for type names.
    """

    def __init__(self):
        """Initialize the registry with an empty dictionary."""
        self._registry: dict[str, type[T]] = {}

    @staticmethod
    def _clean_type_name(type_name: str | Enum) -> str:
        """Convert the type name to uppercase."""
        if isinstance(type_name, Enum):
            return type_name.name.upper()
        return type_name.upper()

    def register(self, type_name: str | Enum, item: type[T]) -> None:
        """
        Register a class with a specified type name (case-insensitive).

        If the type name is already registered, an assertion error is raised.

        Args:
            type_name:
                The type name to be associated with the class.

                The type_name is case-insensitive. If the type_name is an Enum, the name
                (`type_name.name`) is used.
            item: The class to be registered.
        """
        type_name = self._clean_type_name(type_name)
        assert type_name not in self._registry, f"Type '{type_name}' already registered."
        item._type_name = type_name
        self._registry[type_name] = item

    def get(self, type_name: str | Enum) -> type[T]:
        """
        Get the class associated with the given type name.

        Args:
            type_name: The type name of the class to retrieve.
        """
        type_name = self._clean_type_name(type_name)
        return self._registry[type_name]

    def __contains__(self, type_name: str | Enum) -> bool:
        """
        Check if a type name is registered in the registry (case insensitive).

        Args:
            type_name: The type name to check.
        """
        type_name = self._clean_type_name(type_name)
        return type_name in self._registry

    def create_instance(self, type_name: str | Enum, **data: dict) -> T:
        """
        Create an instance of the class associated with the given type name.

        Args:
            type_name (str): The type name of the class to instantiate.
            data: Keyword arguments to be passed to the class constructor.

        Raises:
            ValueError: If the type name is not registered in the registry.
        """
        if isinstance(type_name, Enum):
            type_name = type_name.name
        if type_name.upper() not in self._registry:
            raise ValueError(f"Unknown type `{type_name}`")
        return self._registry[type_name.upper()](**data)



"""Helper functions for OpenAI API."""
from functools import cache
import json
import os
import time
from collections.abc import AsyncGenerator
from openai import AsyncOpenAI
import tiktoken
from tiktoken import Encoding
from sik_llms.models_base import (
    Function,
    FunctionCallResponse,
    FunctionCallResult,
    Client,
    ChatChunkResponse,
    ChatResponseSummary,
    RegisteredClients,
    StructuredOutputResponse,
)

CHAT_MODEL_COST_PER_TOKEN = {
    # minor versions
    'gpt-4o-2024-05-13': {'input': 5.00 / 1_000_000, 'output': 15.00 / 1_000_000},
    'gpt-4o-2024-08-06': {'input': 2.50 / 1_000_000, 'output': 10.00 / 1_000_000},
    'gpt-4o-2024-11-20': {'input': 2.50 / 1_000_000, 'output': 10.00 / 1_000_000},
    'gpt-4o-mini-2024-07-18':  {'input': 0.15 / 1_000_000, 'output': 0.60 / 1_000_000},
    # LEGACY MODELS
    'gpt-4-turbo': {'input': 10.00 / 1_000_000, 'output': 30.00 / 1_000_000},
    'gpt-4-turbo-2024-04-09': {'input': 10.00 / 1_000_000, 'output': 30.00 / 1_000_000},
    'gpt-4-0125-preview': {'input': 0.01 / 1_000, 'output': 0.03 / 1_000},
    'gpt-3.5-turbo': {'input': 0.50 / 1_000_000, 'output': 1.50 / 1_000_000},
    'gpt-3.5-turbo-0125': {'input': 0.50 / 1_000_000, 'output': 1.50 / 1_000_000},
    'gpt-4-0613': {'input': 0.03 / 1_000, 'output': 0.06 / 1_000},
}
CHAT_MODEL_COST_PER_TOKEN_PRIMARY = {
    'gpt-4o-mini': CHAT_MODEL_COST_PER_TOKEN['gpt-4o-mini-2024-07-18'],
    'gpt-4o': CHAT_MODEL_COST_PER_TOKEN['gpt-4o-2024-11-20'],
}
CHAT_MODEL_COST_PER_TOKEN.update(CHAT_MODEL_COST_PER_TOKEN_PRIMARY)


EMBEDDING_MODEL_COST_PER_TOKEN = {
    # "Prices are per 1,000 tokens. You can think of tokens as pieces of words, where 1,000 tokens
    # is about 750 words. This paragraph is 35 tokens."
    # https://openai.com/pricing
    # https://platform.openai.com/docs/models
    ####
    # Embedding models
    ####
    # LATEST MODELS
    # https://openai.com/blog/new-embedding-models-and-api-updates
    'text-embedding-3-small': 0.02 / 1_000_000,
    'text-embedding-3-large': 0.13 / 1_000_000,
    # LEGACY MODELS
    'text-embedding-ada-002': 0.1 / 1_000_000,
}

MODEL_COST_PER_TOKEN = CHAT_MODEL_COST_PER_TOKEN | EMBEDDING_MODEL_COST_PER_TOKEN


@cache
def _get_encoding_for_model(model_name: str) -> Encoding:
    """Gets the encoding for a given model so that we can calculate the number of tokens."""
    return tiktoken.encoding_for_model(model_name)


def num_tokens(model_name: str, value: str) -> int:
    """For a given model, returns the number of tokens based on the str `value`."""
    return len(_get_encoding_for_model(model_name=model_name).encode(value))


def num_tokens_from_messages(model_name: str, messages: list[dict]) -> int:
    """
    Copied from https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    Returns the number of tokens used by a list of messages.
    """
    if model_name in {
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
        # todo: verify once .ipynb is updated
        "gpt-4-1106-preview",
        "gpt-3.5-turbo-1106",
        }:
        tokens_per_message = 3
        tokens_per_name = 1
    elif model_name == "gpt-3.5-turbo-0301":
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-3.5-turbo" in model_name:
        # Warning: gpt-3.5-turbo may update over time.
        # Returning num tokens assuming gpt-3.5-turbo-0613
        return num_tokens_from_messages(model_name="gpt-3.5-turbo-0613", messages=messages)
    elif "gpt-4" in model_name:
        # Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.
        return num_tokens_from_messages(model_name="gpt-4-0613", messages=messages)
    else:
        raise NotImplementedError(f"""num_tokens_from_messages() is not implemented for model {model_name}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens.""")  # noqa
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(_get_encoding_for_model(model_name=model_name).encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def _parse_completion_chunk(chunk) -> ChatChunkResponse:  # noqa: ANN001
    assert chunk.object == 'chat.completion.chunk'
    log_prob = None
    if chunk.choices[0].logprobs:
        log_prob = chunk.choices[0].logprobs.content[0].logprob
    return ChatChunkResponse(
        content=chunk.choices[0].delta.content,
        logprob=log_prob,
    )


@Client.register(RegisteredClients.OPENAI)
class OpenAI(Client):
    """
    Wrapper for OpenAI API which provides a simple interface for calling the
    chat.completions.create method and parsing the response.

    The user can specify the model name, timeout, stream, and other parameters for the API call
    either in the constructor or when calling the object. If the latter, the parameters specified
    when calling the object will override the parameters specified in the constructor.
    """

    def __init__(
            self,
            model_name: str,
            server_url: str | None = None,
            **model_kwargs: dict,
            ) -> None:
        """
        Initialize the wrapper.

        Args:
            client:
                An instance of the AsyncOpenAI client.
            model_name:
                The model name to use for the API call (e.g. 'gpt-4o-mini').
            server_url:
                The base URL for the API call. Required for the `openai-compatible-server` model.
            **model_kwargs: Additional parameters to pass to the API call
        """
        if model_name == 'openai-compatible-server':
            if not server_url:
                raise ValueError("Missing `server_url` for model `openai-compatible-server`")
            api_key = 'None'
            # If max_tokens is not provided, we'll set it;
            # otherwise the server will just output 1 token; not sure if this is llama.cpp issue
            # or something else
            if model_kwargs.get('max_tokens') is None:
                model_kwargs['max_tokens'] = -1
        else:
            server_url = None
            api_key = os.getenv('OPENAI_API_KEY')
            if not api_key:
                raise ValueError(f"Missing `api_key` for model `{model_name}`")

        self.client = AsyncOpenAI(base_url=server_url, api_key=api_key)
        self.model = model_name
        self.model_parameters = model_kwargs or {}

    async def run_async(
        self,
        messages: list[dict],
        model_name: str | None = None,
        **model_kwargs: dict,
    ) -> AsyncGenerator[ChatChunkResponse | ChatResponseSummary | None]:
        """
        Streams chat chunks and returns a final summary. Note that any parameters passed to this
        method will override the parameters passed to the constructor.
        """
        model_name = model_name or self.model
        model_parameters = {**self.model_parameters, **model_kwargs}

        if 'response_format' in model_parameters:
            start_time = time.time()
            completion = await self.client.beta.chat.completions.parse(
                model=model_name,
                messages=messages,
                **model_parameters,
            )
            end_time = time.time()
            yield ChatResponseSummary(
                content=StructuredOutputResponse(
                    parsed=completion.choices[0].message.parsed,
                    refusal=completion.choices[0].message.refusal,
                ),
                total_input_tokens=completion.usage.prompt_tokens,
                total_output_tokens=completion.usage.completion_tokens,
                total_input_cost=completion.usage.prompt_tokens * MODEL_COST_PER_TOKEN[model_name]['input'],  # noqa: E501
                total_output_cost=completion.usage.completion_tokens * MODEL_COST_PER_TOKEN[model_name]['output'],  # noqa: E501
                duration_seconds=end_time - start_time,
            )
        else:
            chunks = []
            start_time = time.time()
            response = await self.client.chat.completions.create(
                model=model_name,
                messages=messages,
                stream=True,
                logprobs=True,
                **model_parameters,
            )
            async for chunk in response:
                if chunk.choices and chunk.choices[0].delta.content:
                    parsed_chunk = _parse_completion_chunk(chunk)
                    yield parsed_chunk
                    chunks.append(parsed_chunk)
            end_time = time.time()
            if model_name == 'openai-compatible-server':
                input_tokens = len(str(messages)) // 4
                output_tokens = sum(len(chunk.content) for chunk in chunks) // 4
                total_input_cost=0
                total_output_cost=0
            else:
                input_tokens = num_tokens_from_messages(model_name, messages)
                output_tokens = sum(num_tokens(model_name, chunk.content) for chunk in chunks)
                total_input_cost=input_tokens * MODEL_COST_PER_TOKEN[model_name]['input']
                total_output_cost=output_tokens * MODEL_COST_PER_TOKEN[model_name]['output']
            yield ChatResponseSummary(
                content=''.join([chunk.content for chunk in chunks]),
                total_input_tokens=input_tokens,
                total_output_tokens=output_tokens,
                total_input_cost=total_input_cost,
                total_output_cost=total_output_cost,
                duration_seconds=end_time - start_time,
            )


@Client.register(RegisteredClients.OPENAI_FUNCTIONS)
class OpenAIFunctions(Client):
    """Wrapper for OpenAI API function calling."""

    def __init__(
            self,
            model_name: str,
            server_url: str | None = None,
            functions: list[Function] | None = None,
            **model_kwargs: dict,
            ) -> None:
        """
        Initialize the wrapper.

        Args:
            client:
                An instance of the AsyncOpenAI client.
            model_name:
                The model name to use for the API call (e.g. 'gpt-4').
            server_url:
                The base URL for the API call. Required for the `openai-compatible-server` model.
            functions:
                List of Function objects defining available functions.
            **model_kwargs:
                Additional parameters to pass to the API call
        """
        if model_name == 'openai-compatible-server':
            if not server_url:
                raise ValueError("Missing `server_url` for model `openai-compatible-server`")
            api_key = 'None'
        else:
            server_url = None
            api_key = os.getenv('OPENAI_API_KEY')
            if not api_key:
                raise ValueError(f"Missing `api_key` for model `{model_name}`")

        self.client = AsyncOpenAI(base_url=server_url, api_key=api_key)
        self.model = model_name
        self.functions = functions or []
        self.model_parameters = model_kwargs or {}
        if 'temperature' not in self.model_parameters:
            self.model_parameters['temperature'] = 0.2


    async def run_async(
            self,
            messages: list[dict[str, str]],
            functions: list[Function] | None = None,
            tool_choice: str = 'required',
            model_name: str | None = None,
            **model_kwargs: dict[str, object],
        ) -> FunctionCallResponse:
        """
        Call the model with functions.

        Args:
            messages: List of messages to send to the model.
            model_name: Optional model override.
            functions: Optional functions override.
            tool_choice:
                Controls which (if any) tool is called by the model. `none` means the model will
                not call any tool and instead generates a message. `auto` means the model can
                pick between generating a message or calling one or more tools. `required` means
                the model must call one or more tools. Specifying a particular tool via
                `{"type": "function", "function": {"name": "my_function"}}` forces the model to
                call that tool.

                `none` is the default when no tools are present. `auto` is the default if tools
                are present.
            **model_kwargs: Additional parameters to override defaults.
        """
        model_name = model_name or self.model
        functions_list = functions or self.functions
        merged_kwargs = {**self.model_parameters, **model_kwargs}
        tools = [func.to_openai_schema() for func in functions_list]

        completion = await self.client.chat.completions.create(
            model=model_name,
            messages=messages,
            tools=tools,
            tool_choice=tool_choice,
            **merged_kwargs,
        )

        # Extract the function call details
        tool_call = completion.choices[0].message.tool_calls[0]
        function_call = FunctionCallResult(
            name=tool_call.function.name,
            arguments=json.loads(tool_call.function.arguments),
            call_id=tool_call.id,
        )

        # Calculate costs
        input_tokens = completion.usage.prompt_tokens
        output_tokens = completion.usage.completion_tokens

        input_cost = input_tokens * CHAT_MODEL_COST_PER_TOKEN[model_name]['input']
        output_cost = output_tokens * CHAT_MODEL_COST_PER_TOKEN[model_name]['output']

        return FunctionCallResponse(
            function_call=function_call,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            input_cost=input_cost,
            output_cost=output_cost,
        )



"""Test the Anthropic Wrapper."""
import asyncio
import os
import pytest
from dotenv import load_dotenv
from sik_llms.models_base import Client, system_message, user_message, RegisteredClients
from sik_llms.anthropic import (
    Anthropic,
    ChatChunkResponse,
    ChatResponseSummary,
)
load_dotenv()

ANTHROPIC_TEST_MODEL = 'claude-3-5-haiku-latest'


@pytest.mark.skipif(os.getenv('ANTHROPIC_API_KEY') is None, reason="ANTHROPIC_API_KEY is not set")
@pytest.mark.asyncio
async def test__async_anthropic_completion_wrapper_call():
    """Test the Anthropic wrapper with multiple concurrent calls."""
    # Create an instance of the wrapper
    client = Anthropic(
        model_name=ANTHROPIC_TEST_MODEL,
        max_tokens=100,
    )
    messages = [
        system_message("You are a helpful assistant."),
        user_message("What is the capital of France?"),
    ]

    async def run_model():  # noqa: ANN202
        chunks = []
        summary = None
        try:
            async for response in client.run_async(messages=messages):
                if isinstance(response, ChatChunkResponse):
                    chunks.append(response)
                elif isinstance(response, ChatResponseSummary):
                    summary = response
            return chunks, summary
        except Exception:
            return [], None

    results = await asyncio.gather(*(run_model() for _ in range(10)))
    passed_tests = []

    for chunks, summary in results:
        response = ''.join([chunk.content for chunk in chunks])
        passed_tests.append(
            'Paris' in response
            and isinstance(summary, ChatResponseSummary)
            and summary.total_input_tokens > 0
            and summary.total_output_tokens > 0
            and summary.total_input_cost > 0
            and summary.total_output_cost > 0
            and summary.duration_seconds > 0,
        )

    assert sum(passed_tests) / len(passed_tests) >= 0.9, (
        f"Only {sum(passed_tests)} out of {len(passed_tests)} tests passed."
    )


def test__Anthropic_registration():
    assert Client.is_registered(RegisteredClients.ANTHROPIC)


@pytest.mark.skipif(os.getenv('ANTHROPIC_API_KEY') is None, reason="ANTHROPIC_API_KEY is not set")
def test__Anthropic__instantiate():
    model = Client.instantiate(
        client_type=RegisteredClients.ANTHROPIC,
        model_name=ANTHROPIC_TEST_MODEL,
    )
    assert isinstance(model, Anthropic)
    assert model.model == ANTHROPIC_TEST_MODEL
    assert model.client is not None
    assert model.client.api_key
    assert model.client.api_key != 'None'


@pytest.mark.skipif(os.getenv('ANTHROPIC_API_KEY') is None, reason="ANTHROPIC_API_KEY is not set")
@pytest.mark.asyncio
async def test__Anthropic__instantiate__run_async():
    model = Client.instantiate(
        client_type=RegisteredClients.ANTHROPIC,
        model_name=ANTHROPIC_TEST_MODEL,
    )
    responses = []
    async for response in model.run_async(messages=[user_message("What is the capital of France?")]):  # noqa: E501
        if isinstance(response, ChatChunkResponse):
            responses.append(response)

    assert len(responses) > 0
    assert 'Paris' in ''.join([response.content for response in responses])



@pytest.mark.skipif(os.getenv('ANTHROPIC_API_KEY') is None, reason="ANTHROPIC_API_KEY is not set")
def test__Anthropic_instantiate___parameters():
    model = Client.instantiate(
        client_type=RegisteredClients.ANTHROPIC,
        model_name=ANTHROPIC_TEST_MODEL,
        temperature=0.5,
        max_tokens=100,
    )
    assert isinstance(model, Anthropic)
    assert model.model == ANTHROPIC_TEST_MODEL
    assert model.model_parameters['temperature'] == 0.5
    assert model.model_parameters['max_tokens'] == 100



"""Helper functions for interacting with the Anthropic API."""
import os
import time
from collections.abc import AsyncGenerator
from anthropic import AsyncAnthropic, Anthropic as SyncAnthropic
from sik_llms.models_base import (
    Client,
    ChatChunkResponse,
    ChatResponseSummary,
    RegisteredClients,
)


CHAT_MODEL_COST_PER_TOKEN = {
    'claude-3-7-sonnet-20250219': {'input': 3.00 / 1_000_000, 'output': 15.00 / 1_000_000},

    'claude-3-5-haiku-20241022': {'input': 0.80 / 1_000_000, 'output': 4.0 / 1_000_000},
    'claude-3-5-sonnet-20241022': {'input': 3.00 / 1_000_000, 'output': 15.00 / 1_000_000},

    'claude-3-opus-20240229': {'input': 15.00 / 1_000_000, 'output': 75.00 / 1_000_000},
}
CHAT_MODEL_COST_PER_TOKEN_LATEST = {
    'claude-3-7-sonnet-latest': CHAT_MODEL_COST_PER_TOKEN['claude-3-7-sonnet-20250219'],

    'claude-3-5-haiku-latest': CHAT_MODEL_COST_PER_TOKEN['claude-3-5-haiku-20241022'],
    'claude-3-5-sonnet-latest': CHAT_MODEL_COST_PER_TOKEN['claude-3-5-sonnet-20241022'],

    'claude-3-opus-latest': CHAT_MODEL_COST_PER_TOKEN['claude-3-opus-20240229'],
}
CHAT_MODEL_COST_PER_TOKEN.update(CHAT_MODEL_COST_PER_TOKEN_LATEST)



def num_tokens(model_name: str, content: str) -> int:
    """Returns the number of tokens for a given string."""
    client = SyncAnthropic()
    response = client.messages.count_tokens(
        model=model_name,
        messages=[{"role": "user", "content": content}],
    )
    return response.tokens

def num_tokens_from_messages(model_name: str, messages: list[dict]) -> int:
    """Returns the number of tokens for a list of messages."""
    client = SyncAnthropic()
    response = client.messages.count_tokens(
        model=model_name,
        messages=messages,
    )
    return response.tokens


def _parse_completion_chunk(chunk) -> ChatChunkResponse | None:  # noqa: ANN001
    """Parse a chunk from the Anthropic API."""
    # Different chunk types in Anthropic's streaming response
    if hasattr(chunk, 'type'):
        if chunk.type == 'content_block_start':
            return None
        if chunk.type == 'content_block_delta':
            return ChatChunkResponse(content=chunk.delta.text)
        if chunk.type == 'message_start':
            return None
        if chunk.type == 'message_delta':
            return ChatChunkResponse(
                content=chunk.delta.text if hasattr(chunk.delta, 'text') else "",
            )
    return None


@Client.register(RegisteredClients.ANTHROPIC)
class Anthropic(Client):
    """
    Wrapper for Anthropic API which provides a simple interface for calling the
    messages.create method and parsing the response.
    """

    def __init__(
            self,
            model_name: str,
            max_tokens: int = 1_000,
            **model_kwargs: dict,
    ) -> None:
        api_key = os.getenv('ANTHROPIC_API_KEY')
        if not api_key:
            raise ValueError("ANTHROPIC_API_KEY is not set")
        self.client = AsyncAnthropic(api_key=api_key)
        self.model = model_name
        if not max_tokens:
            max_tokens = 1_000
        self.model_parameters = {'max_tokens': max_tokens, **model_kwargs}
        # remove any None values
        self.model_parameters = {k: v for k, v in self.model_parameters.items() if v is not None}

    def _convert_messages(self, messages: list[dict]) -> tuple[str, list[dict]]:
        """Convert OpenAI-style messages to Anthropic format."""
        system_content = None
        anthropic_messages = []

        for msg in messages:
            if msg['role'] == 'system':
                system_content = msg['content']
            else:
                anthropic_messages.append({'role': msg['role'], 'content': msg['content']})
        return system_content, anthropic_messages

    async def run_async(
        self,
        messages: list[dict],
        model_name: str | None = None,
        **model_kwargs: dict,
        ) -> AsyncGenerator[ChatChunkResponse | ChatResponseSummary, None]:
        """
        Streams chat chunks and returns a final summary. Parameters passed here
        override those passed to the constructor.
        """
        model_name = model_name or self.model
        model_parameters = {**self.model_parameters, **model_kwargs}
        system_content, anthropic_messages = self._convert_messages(messages)
        api_params = {
            'model': model_name,
            'messages': anthropic_messages,
            'stream': True,
            **model_parameters,
        }
        if system_content:
            api_params['system'] = system_content
        start_time = time.time()
        input_tokens = 0
        output_tokens = 0

        chunks = []
        response = await self.client.messages.create(**api_params)
        async for chunk in response:
            if chunk.type == 'message_start':
                input_tokens = chunk.message.usage.input_tokens
            elif chunk.type == 'message_delta' and hasattr(chunk, 'usage'):
                output_tokens = chunk.usage.output_tokens
            parsed_chunk = _parse_completion_chunk(chunk)
            if parsed_chunk and parsed_chunk.content:
                yield parsed_chunk
                chunks.append(parsed_chunk)
        end_time = time.time()

        yield ChatResponseSummary(
            content=''.join([chunk.content for chunk in chunks]),
            total_input_tokens=input_tokens,
            total_output_tokens=output_tokens,
            total_input_cost=input_tokens * CHAT_MODEL_COST_PER_TOKEN[model_name]['input'],
            total_output_cost=output_tokens * CHAT_MODEL_COST_PER_TOKEN[model_name]['output'],
            duration_seconds=end_time - start_time,
        )
